/* IBM Confidential
 * OCO Source Materials
 * 5724-M24
 * (C) Copyright IBM Corporation 2006, 2007
 * The source code for this program is not published or otherwise
 * divested of its trade secrets, irrespective of what has been
 * deposited with the U.S. Copyright Office.
 */
package com.ibm.wbimonitor.xml.model.mm.util;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.IConfigurationElement;
import org.eclipse.core.runtime.IStatus;
import org.eclipse.core.runtime.Platform;
import org.eclipse.emf.common.util.URI;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.resource.Resource;
import org.eclipse.emf.ecore.xml.type.internal.QName;

import com.ibm.wbimonitor.xml.model.eventdefinition501.DocumentRoot;
import com.ibm.wbimonitor.xml.model.mm.EventPartType;
import com.ibm.wbimonitor.xml.model.mm.ImportType;
import com.ibm.wbimonitor.xml.model.mm.MonitorType;
import com.ibm.wbimonitor.xml.model.util.Logger;
import com.ibm.wbimonitor.xml.model.util.MMHelper;
import com.ibm.wbimonitor.xml.model.util.URIAdapterUtil;


public class EventTypeResolver {
	public static final String COPYRIGHT="(C) Copyright IBM Corporation 2006, 2007.";  //$NON-NLS-1$
	public static final String SMO_SCHEME = "smo://";
	private static List<IEventTypeResolver> sEventTypeResolvers = null;

	/**
	 * Finds and resolves the given event inside a CBE file imported into the monitor model
	 * 
	 * @param parent An EObject in the monitor model that contains the appropriate event model import
	 * @param eventName The event name to look for amongst the imported .cbe file.
	 * @return The resolved EventDefinitionType, or null if a type with the given name could not be found.
	 */
	public static EObject resolve(EObject parent, String eventName){
		
		/* Translate the typePath into the actual EObject to get.  
		 * We don't know what import provides the event type in question, so keep resolving them 
		 * until we find the one that contains the event we're interested in */
		
		/* NOTE: This method requires that 'parent' be within a Monitor container already */
		MonitorType monType = MMHelper.getMonitorType(parent);
		if(monType == null || monType.getEventModel() == null ) {
		    return null;
		}
		
		for (ImportType eventImport : (List<ImportType>)monType.getEventModel().getImport()){
			if(eventImport.getLocation() != null && eventImport.getLocation().endsWith(Constants.eventModelCbeExt)){
				/* Get the events within this import */
				List eventList = eventLoader(eventImport);
				if (eventList == null){
					continue;
				}
				com.ibm.wbimonitor.xml.model.eventdefinition501.EventDefinitionType curEventType = null;
				for (Iterator it2 = eventList.iterator(); it2.hasNext(); ){
					curEventType = (com.ibm.wbimonitor.xml.model.eventdefinition501.EventDefinitionType)it2.next();
					if (curEventType.getName().equals(eventName)){
						return curEventType;
					}
				}
			}			
		}
		return null;
	}
	
	/**
	 * This method resolves a QName reference from an EventPart to the actual XSDTypeDefinition
	 * object that is referenced.
	 * 
	 * @param parent The EventPartType object that references the type definition
	 * @param eventName
	 * @return
	 */
	public static EObject resolve(EventPartType parent, QName typeQName){
		/* Iterate through the contributed resolvers looking for the event definition */
		EObject xsdTypeDef = null;
		for (IEventTypeResolver resolver : getResolvers()){
			xsdTypeDef = resolver.resolve(parent, typeQName);
			if(xsdTypeDef != null){
				return xsdTypeDef;
			}
		}		
		
		return null;
	}

	/**
	 * Loads the IEventTypeResolvers that have been contributed by Eclipse extension, if there are any
	 * and caches the information so these resolvers do not need to be reloaded each time.
	 * 
	 * @return The list of resolved IEventTypeResolvers contributed via extension
	 */
	private static List<IEventTypeResolver> getResolvers(){
		if (sEventTypeResolvers == null){
			sEventTypeResolvers = new ArrayList<IEventTypeResolver>();
			IConfigurationElement[] eventTypeResolvers = Platform.getExtensionRegistry().getConfigurationElementsFor("com.ibm.wbimonitor.xml.model.eventTypeResolver");
			if (eventTypeResolvers != null){
				for(int i = 0; i < eventTypeResolvers.length; i++){
					try {
						sEventTypeResolvers.add((IEventTypeResolver)eventTypeResolvers[i].createExecutableExtension("class"));
					} catch (CoreException e){
						Logger.log(IStatus.ERROR, "Failed to create event type resolver", e);
					}
				}
			}
		}

		return sEventTypeResolvers;
	}
	
	
	/**
	 * Loads the list of CBE EventDefinitionType objects pointed by by the given event model import
	 * 
	 * @param importType The event model import to resolve
	 * @return The list of EventDefinitionType objects pointed at by the given import
	 */
	public static List eventLoader(ImportType importType){
		/* Search for an existing resource with this URI in the resource set (if there is one) */
		if (importType.eResource() != null && importType.eResource().getResourceSet() != null && importType.getLocation() != null){
			try {
				URI mmUri = importType.eResource().getURI();
				URI eventDefURI = URIAdapterUtil.resolveImportUri(mmUri, importType.getLocation());
				if (URIAdapterUtil.exists(eventDefURI)) {
					Resource res = importType.eResource().getResourceSet().getResource(eventDefURI, true);
					DocumentRoot eventDocRoot = (DocumentRoot)res.getContents().get(0);
					if (eventDocRoot.getEventDefinitionList() != null){
						return eventDocRoot.getEventDefinitionList().getEventDefinition();
					}
				}
			} catch (IllegalArgumentException cause) {
				// Invalid user input, ignore this error and return null.
			}
		}
		
		return null;
	}
	
	/**
	 * Returns the target of the given event model import.  If the importType is referencing an XSD type 
	 * (either in the XML catalog or somewhere specifically on disk), a list of XSDSchema objects is returned.
	 * (Note that this behaviour is contributed via extension and may or may not be present).
	 * 
	 * If the importType is referencing a .cbe file, a list of EventDefinitionType objects is returned.
	 * 
	 * @param importType The importType to resolve
	 * @return The Object (or list of Objects) pointed to by the given import 
	 */
	public static Object resolveImport(ImportType importType){
		for (IEventTypeResolver resolver : getResolvers()){
			if (resolver.isApplicableForImport(importType)){
				return resolver.resolveImport(importType);
			}
		}
		
		/* If there are no applicable resolvers, treat it as a CBE case */
		return eventLoader(importType);
	}
	

}
