/* IBM Confidential
 * OCO Source Materials
 * 5724-M24
 * (C) Copyright IBM Corporation 2006, 2007
 * The source code for this program is not published or otherwise
 * divested of its trade secrets, irrespective of what has been
 * deposited with the U.S. Copyright Office.
 */
package com.ibm.wbimonitor.xml.model.util;

import java.io.File;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

import org.eclipse.core.resources.IFile;
import org.eclipse.core.resources.IProject;
import org.eclipse.core.resources.IResource;
import org.eclipse.core.resources.IResourceVisitor;
import org.eclipse.core.resources.ResourcesPlugin;
import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.IPath;
import org.eclipse.core.runtime.IStatus;
import org.eclipse.core.runtime.Path;
import org.eclipse.emf.common.util.URI;
import org.eclipse.emf.ecore.resource.Resource;
import org.eclipse.emf.ecore.resource.ResourceSet;

import com.ibm.wbimonitor.xml.model.ModelPlugin;
import com.ibm.wbimonitor.xml.model.eventdefinition501.DocumentRoot;
import com.ibm.wbimonitor.xml.model.eventdefinition501.EventDefinitionListType;
import com.ibm.wbimonitor.xml.model.eventdefinition501.EventDefinitionType;
import com.ibm.wbimonitor.xml.model.mm.util.Constants;

/**
 * An utility class that helps to resolve parent event definition.
 * This utility does not work outside of an Eclipse environment.
 */
public class EventDefinitionParentResolver {
public static final String COPYRIGHT="(C) Copyright IBM Corporation 2006, 2007.";  //$NON-NLS-1$



	
	/**
	 * Returns the file extension for a given file descriptor.
	 * @param file A file descriptor.
	 * @return File extension, or null if 
	 */
	private static String getFileExtension(File file) {
		if (file != null) {
			String fileName = file.getName();
			if (fileName != null) {
				int index = fileName.lastIndexOf('.');
				if (index >= 0 && fileName.length() > (index + 1)) {
					String fileExtension = fileName.substring(index + 1);
					return fileExtension;
				}
			}
		}
		return null;
	}
	
	/**
	 * Resolves the parent Event Definition Type for a given event definition type.
	 * precondition: The event definition must be owned by a valid resource, which corresponds
	 * to a file in the file system, within a resource set.
	 * 
	 * Same as resolveParent(eventDefinition, false)
	 * 
	 * @param eventDefinition
	 * @return The parent event definition type if resolved; otherwise null.
	 */
	public static EventDefinitionType resolveParent(EventDefinitionType eventDefinition) {
		return resolveParent(eventDefinition, false);
	}
	
	/**
	 * Resolves the parent Event Definition Type for a given event definition type.
	 * precondition: The event definition must be owned by a valid resource, which corresponds
	 * to a file in the file system, within a resource set.
	 * @param eventDefinition
	 * @return The parent event definition type if resolved; otherwise null.
	 */
	public static EventDefinitionType resolveParent(EventDefinitionType eventDefinition, boolean includeBaseEventDefinition) {
		
		if (eventDefinition != null && eventDefinition.eResource() != null && eventDefinition.eResource().getResourceSet() != null) {
			String parentName = eventDefinition.getParent();
			if (parentName != null) {
				URI uri = eventDefinition.eResource().getURI();
				ResourceSet resourceSet = eventDefinition.eResource().getResourceSet();

				List scope = getSearchScope(uri, includeBaseEventDefinition);
				for (Iterator it = scope.iterator(); it.hasNext();) {
					URI childUri = (URI) it.next();
					EventDefinitionType result = getEventDefinitionType(resourceSet, childUri, parentName);
					if (result != null) {
						return result;
					}								
				}				
			}
		}
		return null;
	}
	
	/**
	 * Returns an EventDefinitionType if a match is found in the given file. The matching
	 * Event definition type's name must be exactly the same as the given name.
	 * 
	 * Calling this method will result in loading the given file into the given resource set if
	 * it has not been loaded.
	 * @param resourceSet A resource set.
	 * @param uri An uri.
	 * @param name Name of event definition type to search for.
	 * @return An EventDefinitionType if a match is found; otherwise null.
	 */
	public static EventDefinitionType getEventDefinitionType(ResourceSet resourceSet, URI uri, String name) {
		
		if (resourceSet != null && name != null) {
			Resource resource = null;
			if (BaseEventDefinitionUtil.BASE_EVENT_DEFINITION_URI.equals(uri)) {
				resource = BaseEventDefinitionUtil.getBaseEventDefinitionResource(resourceSet);
			} else {
				resource = resourceSet.getResource(uri, true);	
			}
			if (resource != null) {
				if (! resource.getContents().isEmpty()) {
					Object root = resource.getContents().get(0);
					if (root instanceof DocumentRoot) {
						EventDefinitionListType eventList = ((DocumentRoot) root).getEventDefinitionList();
						if (eventList != null && eventList.getEventDefinition() != null) {
							for (Iterator it = eventList.getEventDefinition().iterator(); it.hasNext();) {
								EventDefinitionType eventDefinitionType = (EventDefinitionType) it.next();
								if (name.equals(eventDefinitionType.getName())) {
									return eventDefinitionType;
								}
							}
						}
					}
				}
			}
		}
		return null;
	}
	
	/**
	 * This method returns a list of URIs that are files in the same folder as the given URI
	 * with CBE extension.
	 * @param uri
	 * @return
	 */
	private static List getSearchScope(URI uri, boolean includeBaseEventDefinition) {
		final List result = new ArrayList();
		if (uri.isFile()) {
			// File URI
			String path = uri.toFileString();
			File file = new File(path);
			File parent = file.getParentFile();
			if (parent != null && parent.exists()) {
				File[] children = parent.listFiles();
				for (int i = 0; i < children.length; i++) {
					File child = children[i];
					if (child.isFile() && child.exists() && Constants.eventModelCbeExt.equalsIgnoreCase(getFileExtension(child))) {
						URI childUri = URI.createFileURI(child.getAbsolutePath());
						result.add(childUri);				
					}
				}
			}

		} else if ("platform".equals(uri.scheme()) && uri.segmentCount() > 1 && "resource".equals(uri.segment(0))) {
			// Platform URI
			String path = URIAdapterUtil.decodeURI(uri);
			IFile file = ResourcesPlugin.getWorkspace().getRoot().getFile(new Path(path));
			if (file != null && file.exists()) {
			
				IResourceVisitor visitor = new IResourceVisitor() {
					public boolean visit(IResource resource) throws CoreException {
						if (resource instanceof IFile) {
							if (resource.exists() && Constants.eventModelCbeExt.equalsIgnoreCase(resource.getFileExtension())) {
								URI childUri = URIAdapterUtil.encodePlatformResourceURI(resource.getFullPath().toString());
								result.add(childUri);
							}
						}
						return true;
					}
				};
					IProject project = file.getProject();
					List projects = getReferencedProjects(project);
					for (Iterator it = projects.iterator(); it.hasNext();) {
						IProject theProject = (IProject) it.next();
						try {
							theProject.accept(visitor);	
						} catch (CoreException cause) {
							Logger.log(IStatus.ERROR, "An error occurred while collecting the event search scope for URI '" + uri.toString() + "'.", cause);
						}
					}					

					
			}			
		}
		if (includeBaseEventDefinition) {
			result.add(BaseEventDefinitionUtil.getBaseEventDefinitionURI());
		}
		return result;
	}
	
	
	private static List getReferencedProjects(IProject project) {
		List result = new ArrayList();
		List current = new ArrayList();
		current.add(project);
		while (! current.isEmpty()) {
			IProject currentProject = (IProject) current.remove(0);
			if (! result.contains(currentProject)) {
				result.add(currentProject);
				try {
					IProject[] referencedProjects = currentProject.getReferencedProjects();
					for (int i = 0; i < referencedProjects.length; i++) {
						current.add(referencedProjects[i]);
					}
				} catch (CoreException cause) {
					ModelPlugin.getDefault().getLog().log(cause.getStatus());
				}
			}
		}
		return result;
	}
	
	private static String getEncodedFullPath(IResource file) {
		IPath path = file.getFullPath();
		String[] segments = path.segments();
		StringBuffer buffer = new StringBuffer();
		for (int i = 0; i < segments.length; i++) {
			buffer.append('/');
			//buffer.append(URI.encodeSegment(segments[i], false));
			// Do not encode segments.
			buffer.append(segments[i]);
		}
		
		return buffer.toString();
	}
	
	/* (non-Javadoc)
	 * @see com.ibm.wbimonitor.xml.model.eventdefinition501.EventDefinitionType#getAllExtendedDataElements()
	 */
	public static List getAllExtendedDataElements(EventDefinitionType eventDefType) {
		return getAllExtendedDataElements(eventDefType,true,true);
		
	}
	
	public static List getAllExtendedDataElements(EventDefinitionType eventDefType, boolean includeBaseEventDefinition, boolean includeCBE) {
		List result = new ArrayList();
		if(eventDefType.getName().equals(Constants.CommonBaseEventName)){
			if(includeCBE){
				result.addAll(eventDefType.getExtendedDataElement());
			}
		}else{
			result.addAll(eventDefType.getExtendedDataElement());
		}
		EventDefinitionType parent = EventDefinitionParentResolver.resolveParent(eventDefType, includeBaseEventDefinition);
		if (parent != null) {
			result.addAll(getAllExtendedDataElements(parent,includeBaseEventDefinition, includeCBE));
		}
		return result;
	}
	/* (non-Javadoc)
	 * @see com.ibm.wbimonitor.xml.model.eventdefinition501.EventDefinitionType#getAllProperties()
	 */
	public static List getAllProperties(EventDefinitionType eventDefType) {
		return getAllProperties(eventDefType, true, true);
	}
	
	public static List getAllProperties(EventDefinitionType eventDefType, boolean includeBaseEventDefinition, boolean includeCBE) {
		List result = new ArrayList();
		if(eventDefType.getName().equals(Constants.CommonBaseEventName)){
			if(includeCBE){
				result.addAll(eventDefType.getProperty());
			}
		}else{
			result.addAll(eventDefType.getProperty());
		}
		EventDefinitionType parent = EventDefinitionParentResolver.resolveParent(eventDefType, includeBaseEventDefinition);
		if (parent != null) {
			result.addAll(getAllProperties(parent,includeBaseEventDefinition,includeCBE));
		}
		return result;
	}
}
