/* IBM Confidential
 * OCO Source Materials
 * 5724-M24
 * (C) Copyright IBM Corporation 2006, 2007
 * The source code for this program is not published or otherwise
 * divested of its trade secrets, irrespective of what has been
 * deposited with the U.S. Copyright Office.
 */
package com.ibm.wbimonitor.xml.model.util;

import org.apache.xerces.util.XMLChar;

/**
 * This is a singleton class that converts a Java String
 * to and from a valid NCName.  Invalid NCName characters
 * are escaped by _xUUUU_ where UUUU is the Hex digits of
 * the character's Unicode representation.
 */
public class NCNameConverter {
public static final String COPYRIGHT="(C) Copyright IBM Corporation 2006, 2007.";  //$NON-NLS-1$

	
	private final static String UNDERSCORE_CHAR = "_x005f_";
	
	
	/**
	 * Convert a Hex character to a decimal value.
	 * Throws <code>InvalidNCNameException</code> if Hex character
	 * is invalid.
	 * 
	 * @param hex the Hex character to convert
	 * @return the converted decimal value
	 * @throws InvalidNCNameException
	 */
	private static int hexCharToDecimal(char hex) throws InvalidNCNameException {
		
		switch (hex) {
			case '0' : return 0;
			case '1' : return 1;
			case '2' : return 2;
			case '3' : return 3;
			case '4' : return 4;
			case '5' : return 5;
			case '6' : return 6;
			case '7' : return 7;
			case '8' : return 8;
			case '9' : return 9;
			case 'a' : return 10;
			case 'A' : return 10;
			case 'b' : return 11;
			case 'B' : return 11;
			case 'c' : return 12;
			case 'C' : return 12;
			case 'd' : return 13;
			case 'D' : return 13;
			case 'e' : return 14;
			case 'E' : return 14;
			case 'f' : return 15;
			case 'F' : return 15;
		}
		throw new InvalidNCNameException();
	}
	
	
	/**
	 * Convert a decimal value to a Hex character.
	 * 
	 * @param decimal
	 * @return the converted Hex character
	 */
	private static char decimalToHexChar(int decimal) {
		
		switch (decimal) {
			case 0 : return '0';
			case 1 : return '1';
			case 2 : return '2';
			case 3 : return '3';
			case 4 : return '4';
			case 5 : return '5';
			case 6 : return '6';
			case 7 : return '7';
			case 8 : return '8';
			case 9 : return '9';
			case 10 : return 'a';
			case 11 : return 'b';
			case 12 : return 'c';
			case 13 : return 'd';
			case 14 : return 'e';
			case 15 : return 'f';
		}
		return '0';
	}
	
	
	/**
	 * Convert a four-character Hex representation to 
	 * a decimal value.  Assume UCS-2 encoding.
	 * 
	 * @param hex the Hex String
	 * @return the converted decimal value
	 * @throws InvalidNCNameException
	 */
	private static int hexToDecimal(String hex) throws InvalidNCNameException {
		
		if (hex.length() != 4) {
			throw new InvalidNCNameException(); 
		}
		
		// Assume UCS-2 encoding
		int decimal = 0;
		decimal = decimal + (hexCharToDecimal(hex.charAt(0)) * 4096);
		decimal = decimal + (hexCharToDecimal(hex.charAt(1)) * 256);
		decimal = decimal + (hexCharToDecimal(hex.charAt(2)) * 16);
		decimal = decimal + (hexCharToDecimal(hex.charAt(3)));
		
		return decimal;
	}
	
	
	/**
	 * Convert a decimal value to its Hex representation.
	 * Assume UCS-2 encoding.
	 * 
	 * @param decimal the decimal value
	 * @return the Hex representation
	 */
	private static String decimalToHex(int decimal) {
		
		StringBuffer sb = new StringBuffer();
		
		// Assume UCS-2 encoding
		int remainder;
		sb.append(decimalToHexChar(decimal / 4096));
		remainder = decimal % 4096;
		sb.append(decimalToHexChar(remainder / 256));
		remainder = remainder % 256;
		sb.append(decimalToHexChar(remainder / 16));
		remainder = remainder % 16;
		sb.append(decimalToHexChar(remainder));
		
		return sb.toString();
	}
	

	/**
	 * Return a valid NCName character by escaping
	 * the input character.
	 * 
	 * @param c the character to escape
	 * @return the escaped character
	 */
	private static String getNcnameChar(char c) {
		
		StringBuffer sb = new StringBuffer();
		
		sb.append("_x");
		sb.append(decimalToHex(c));
		sb.append("_");
		
		return sb.toString();
	}


	/**
	 * Determine whether the input String is a valid
	 * NCName.
	 * 
	 * @param name
	 * @return wheter the name is valid
	 */
	public static boolean isValidNCName(String name) {
		
		if (name != null) {
			return XMLChar.isValidNCName(name);
		}
		return false;
	}


	/**
	 * Converts a NCName to a Java String, resolving all
	 * escaped characters.  <code>InvalidNCNameException</code>
	 * is thrown if the NCName is invalid.
	 * 
	 * @param ncname the NCName to convert
	 * @return the converted Java String
	 * @throws InvalidNCNameException
	 */
	public static String ncnameToString(String ncname) throws InvalidNCNameException {
	
		if (!isValidNCName(ncname)) {
			throw new InvalidNCNameException();
		}
	
		StringBuffer sb = new StringBuffer();
		
		int stringLen = ncname.length();
		char c;
		int charEndIndex;
		String hex;
		
		for (int i=0; i<stringLen; i++) {
			c = ncname.charAt(i);
			
			if (i < stringLen-1 && c == '_' && (ncname.charAt(i+1) == 'x' || ncname.charAt(i+1) == 'X')) {
					
				charEndIndex = ncname.indexOf('_', i+1);
				hex = ncname.substring(i+2, charEndIndex);
				sb.append((char)hexToDecimal(hex));
				i = charEndIndex;
			} else {
				sb.append(c);
			}
		}
		
		return sb.toString();
	}


	/**
	 * Converts the input String to a proper NCName, escaping
	 * all invalid characters
	 * 
	 * @param string the Java string to convert
	 * @return the converted NCName
	 */
	public static String stringToNcname(String string) {
		
		if (string == null) {
			return null;
		}
		
		StringBuffer sb = new StringBuffer();
		
		int stringLen = string.length();
		char c;
		
		for (int i=0; i<stringLen; i++) {
			c = string.charAt(i);
			if (i < stringLen-1 && c == '_' && (string.charAt(i+1) == 'x' || string.charAt(i+1) == 'X')) {
				sb.append(UNDERSCORE_CHAR);
			} else if (i == 0) {
				if (XMLChar.isNCNameStart(c)) {
					sb.append(c);
				} else {
					sb.append(getNcnameChar(c));
				}
			} else {
				if (XMLChar.isNCName(c)) {
					sb.append(c);
				} else {
					sb.append(getNcnameChar(c));
				}
			}
		}
		
		return sb.toString();
	}

}
