/* IBM Confidential
 * OCO Source Materials
 * 5724-M24
 * (C) Copyright IBM Corporation 2006, 2007
 * The source code for this program is not published or otherwise
 * divested of its trade secrets, irrespective of what has been
 * deposited with the U.S. Copyright Office.
 */
package com.ibm.wbimonitor.xml.model.util;

import java.io.File;
import java.io.InputStream;
import java.net.URL;
import java.net.URLConnection;
import java.util.StringTokenizer;

import org.eclipse.core.resources.IFile;
import org.eclipse.core.resources.ResourcesPlugin;
import org.eclipse.core.runtime.IStatus;
import org.eclipse.core.runtime.Path;
import org.eclipse.emf.common.util.URI;

/**
 * URI Utilities
 */
public class URIAdapterUtil {
public static final String COPYRIGHT="(C) Copyright IBM Corporation 2006, 2007.";  //$NON-NLS-1$

	
	/**
	 * Creates an encoded Platform resource URI, such that characters like 'whitespace'
	 * and '#' are escaped.
	 * 
	 * IMPORTANT: All MME components must use encoded platform resource URI when loading
	 * an EMF resource.
	 * 
	 * @param path An absolute workspace file path. e.g. "/project 1/event folder/myEvent.cbe"
	 * @return Encoded platform resource URI. e.g. "platform:/resource/project%201/event%20folder/myEvent.cbe"
	 */ 
	public static URI encodePlatformResourceURI(String path) {
		String encodedPath = encodePath(path);
		URI uri = URI.createPlatformResourceURI(encodedPath);
		return uri;
	}
	
	/**
	 * Decode a Platform resource URI or a relative URI.
	 * 
	 * NOTE: All MME components must invoke this method to convert an URI to string representation. 
	 * E.g. Setting location attribute for an ImportType.
	 * 
	 * @param uri Encoded platform resource URI. e.g. "platform:/resource/project%201/event%20folder/myEvent.cbe"
	 * @return A decoded string path. e.g. "/project 1/event folder/myEvent.cbe"
	 */
	public static String decodeURI(URI uri) {
		if (uri != null) {
			if ("platform".equals(uri.scheme()) && uri.segmentCount() > 1 && "resource".equals(uri.segment(0))) {
				// Platform URI
				StringBuffer buffer = new StringBuffer();
				int size = uri.segmentCount();
				for (int i = 1; i < size; i++) {
					buffer.append('/');
					buffer.append(URI.decode(uri.segment(i)));
				}
				
				return buffer.toString();
			} else if (uri.isRelative()) {
				StringBuffer buffer = new StringBuffer();
				int size = uri.segmentCount();
				for (int i = 0; i < size; i++) {
					if (i > 0 || uri.toString().startsWith("/")) {
						buffer.append('/');
					}
					buffer.append(URI.decode(uri.segment(i)));
				}
				
				return buffer.toString();
				
			}
		}
		// Return the origianl URI if it's not a platform URI.
		return uri.path();
	}

	/**
	 * 
	 * @param uri Encoded URI
	 * @return
	 */
	public static IFile toIFile(URI uri) {
		IFile result = null;
		if ("platform".equals(uri.scheme()) && uri.segmentCount() > 1 && "resource".equals(uri.segment(0))) {
			// Platform URI
			StringBuffer buffer = new StringBuffer();
			int size = uri.segmentCount();
			for (int i = 1; i < size; i++) {
				buffer.append('/');
				buffer.append(URI.decode(uri.segment(i)));
			}
			result = ResourcesPlugin.getWorkspace().getRoot().getFile(new Path(buffer.toString()));
			
		}
		return result;
	}

	/**
	 * Convert an URI to a system File descriptor.
	 * @param uri ENcoded URI
	 * @return A file descriptor.
	 */
	public static File toFile(URI uri) {
		File file = null;
		if (uri.isFile()) {
			// File URI
			String path = uri.toFileString();
			file = new File(path);
		} else if ("platform".equals(uri.scheme()) && uri.segmentCount() > 1 && "resource".equals(uri.segment(0))) {
			// Platform URI
			StringBuffer buffer = new StringBuffer();
			int size = uri.segmentCount();
			for (int i = 1; i < size; i++) {
				buffer.append('/');
				buffer.append(URI.decode(uri.segment(i)));
			}
			IFile resource = ResourcesPlugin.getWorkspace().getRoot().getFile(new Path(buffer.toString()));
			if(resource != null && resource.getLocation()!=null){
				file = resource.getLocation().toFile();
			}
		}
		
		return file;
	}
	


	/**
	 * An utility that creates an encoded relative URI.
	 */ 
	public static URI encodeRelativeURI(String path) {
		String encodedPath = encodePath(path);
		URI uri = URI.createURI(encodedPath);
		return uri;
	}
	
	/**
	 * creates an encoded file URI
	 * @param path
	 * @return
	 */
	public static URI encodeFileURI(String path){
		String encodedPath = encodePath(path);
		URI uri = URI.createFileURI(encodedPath);
		return uri;
	}

	private static String encodePath(String path) {
		if (path != null && path.length() > 0) {
			StringTokenizer tokenizer = new StringTokenizer(path, "/");
			String[] segments = new String[tokenizer.countTokens()]; 
			int i = 0;
			while (tokenizer.hasMoreTokens()) {
				String token = tokenizer.nextToken();
				segments[i] = URI.encodeSegment(token, false);
				i++;
			}
			StringBuffer buffer = new StringBuffer();
			if (path.startsWith("/")) {
				buffer.append("/");
			}
			for (int j = 0; j < segments.length; j++) {
				if (j > 0) {
					buffer.append("/");
				}
				buffer.append(segments[j]);
			}
			
			return buffer.toString();
		}
		return path;
		
		
	}
	
	
	public static URI resolveImportUri(URI base, String location) {
		URI locationURI = URI.createURI(location);
		if (!locationURI.isRelative()) return locationURI;
		
		if (base.isFile()) {
			// File URI
			return locationURI.resolve(base);
	
		} else if ("platform".equals(base.scheme()) && base.segmentCount() > 1 && "resource".equals(base.segment(0))) {
			try {
				IFile baseFile = URIAdapterUtil.toIFile(base);
				Path path = new Path(location);
				IFile importFile = null;
				if (path.isAbsolute()) {
					importFile = baseFile.getWorkspace().getRoot().getFile(path);
				} else {
					importFile = baseFile.getParent().getFile(path);
				}
				if (importFile != null) {
					
					return URIAdapterUtil.encodePlatformResourceURI(importFile.getFullPath().toString());
				}
			} catch (IllegalArgumentException cause) {
				// Ignore, invalid user input.
			}			
		}
		return null;
	}	

	
	/**
	 * Test whether the file, which is 
	 * @param uri
	 * @return
	 */
	public static boolean exists(URI uri) {
		File file = toFile(uri);
		if (file != null) {
			if (file.exists()) {
				return true;
			}
		} else {
			/* This might be a 'jar:' URI */
			try {
				InputStream is = new URL(uri.toString()).openStream();
				is.close();
				return true;
			} catch (Exception e){
				/* There was something wrong with resolving the URI, so it must not 'exist' */
				Logger.log(IStatus.ERROR, "An error occurred while trying to resolve the URI '" + uri.toString() + "'", e);
				return false;
			}
		}
		return false;
	}
}
